@@ -1,7 +1,10 @@
+from __future__ import annotations
+
 import argparse
+import json
 import operator
 import sys
-from typing import Any, Dict, List, Optional, Type, TYPE_CHECKING, Union
+from typing import Any, TYPE_CHECKING
 
 import gitlab
 import gitlab.base
@@ -16,9 +19,9 @@ class GitlabCLI:
         gl: gitlab.Gitlab,
         gitlab_resource: str,
         resource_action: str,
-        args: Dict[str, str],
+        args: dict[str, str],
     ) -> None:
-        self.cls: Type[gitlab.base.RESTObject] = cli.gitlab_resource_to_cls(
+        self.cls: type[gitlab.base.RESTObject] = cli.gitlab_resource_to_cls(
             gitlab_resource, namespace=gitlab.v4.objects
         )
         self.cls_name = self.cls.__name__
@@ -26,26 +29,17 @@ class GitlabCLI:
         self.resource_action = resource_action.lower()
         self.gl = gl
         self.args = args
-        self.parent_args: Dict[str, Any] = {}
-        self.mgr_cls: Union[
-            Type[gitlab.mixins.CreateMixin],
-            Type[gitlab.mixins.DeleteMixin],
-            Type[gitlab.mixins.GetMixin],
-            Type[gitlab.mixins.GetWithoutIdMixin],
-            Type[gitlab.mixins.ListMixin],
-            Type[gitlab.mixins.UpdateMixin],
-        ] = getattr(gitlab.v4.objects, f"{self.cls.__name__}Manager")
+        self.parent_args: dict[str, Any] = {}
+        self.mgr_cls: Any = getattr(gitlab.v4.objects, f"{self.cls.__name__}Manager")
         # We could do something smart, like splitting the manager name to find
         # parents, build the chain of managers to get to the final object.
         # Instead we do something ugly and efficient: interpolate variables in
         # the class _path attribute, and replace the value with the result.
-        if TYPE_CHECKING:
-            assert self.mgr_cls._path is not None
 
         self._process_from_parent_attrs()
 
         self.mgr_cls._path = self.mgr_cls._path.format(**self.parent_args)
-        self.mgr = self.mgr_cls(gl)
+        self.mgr: Any = self.mgr_cls(gl)
         self.mgr._from_parent_attrs = self.parent_args
         if self.mgr_cls._types:
             for attr_name, type_cls in self.mgr_cls._types.items():
@@ -81,8 +75,10 @@ class GitlabCLI:
         return self.do_custom()
 
     def do_custom(self) -> Any:
-        class_instance: Union[gitlab.base.RESTManager, gitlab.base.RESTObject]
-        in_obj = cli.custom_actions[self.cls_name][self.resource_action][2]
+        class_instance: (
+            gitlab.base.RESTManager[gitlab.base.RESTObject] | gitlab.base.RESTObject
+        )
+        in_obj = cli.custom_actions[self.cls_name][self.resource_action].in_object
 
         # Get the object (lazy), then act
         if in_obj:
@@ -131,25 +127,37 @@ class GitlabCLI:
             assert isinstance(self.mgr, gitlab.mixins.CreateMixin)
         try:
             result = self.mgr.create(self.args)
+            if TYPE_CHECKING:
+                assert isinstance(result, gitlab.base.RESTObject)
         except Exception as e:  # pragma: no cover, cli.die is unit-tested
             cli.die("Impossible to create object", e)
         return result
 
-    def do_list(
-        self,
-    ) -> Union[gitlab.base.RESTObjectList, List[gitlab.base.RESTObject]]:
+    def do_list(self) -> list[gitlab.base.RESTObject]:
         if TYPE_CHECKING:
             assert isinstance(self.mgr, gitlab.mixins.ListMixin)
+        message_details = gitlab.utils.WarnMessageData(
+            message=(
+                "Your query returned {len_items} of {total_items} items. To return all "
+                "items use `--get-all`. To silence this warning use `--no-get-all`."
+            ),
+            show_caller=False,
+        )
+
         try:
-            result = self.mgr.list(**self.args)
+            result = self.mgr.list(
+                **self.args, message_details=message_details, iterator=False
+            )
         except Exception as e:  # pragma: no cover, cli.die is unit-tested
             cli.die("Impossible to list objects", e)
         return result
 
-    def do_get(self) -> Optional[gitlab.base.RESTObject]:
+    def do_get(self) -> gitlab.base.RESTObject | None:
         if isinstance(self.mgr, gitlab.mixins.GetWithoutIdMixin):
             try:
                 result = self.mgr.get(id=None, **self.args)
+                if TYPE_CHECKING:
+                    assert isinstance(result, gitlab.base.RESTObject) or result is None
             except Exception as e:  # pragma: no cover, cli.die is unit-tested
                 cli.die("Impossible to get object", e)
             return result
@@ -161,6 +169,8 @@ class GitlabCLI:
         id = self.args.pop(self.cls._id_attr)
         try:
             result = self.mgr.get(id, lazy=False, **self.args)
+            if TYPE_CHECKING:
+                assert isinstance(result, gitlab.base.RESTObject) or result is None
         except Exception as e:  # pragma: no cover, cli.die is unit-tested
             cli.die("Impossible to get object", e)
         return result
@@ -175,7 +185,7 @@ class GitlabCLI:
         except Exception as e:  # pragma: no cover, cli.die is unit-tested
             cli.die("Impossible to destroy object", e)
 
-    def do_update(self) -> Dict[str, Any]:
+    def do_update(self) -> dict[str, Any]:
         if TYPE_CHECKING:
             assert isinstance(self.mgr, gitlab.mixins.UpdateMixin)
         if issubclass(self.mgr_cls, gitlab.mixins.GetWithoutIdMixin):
@@ -200,19 +210,24 @@ else:
 
 
 def _populate_sub_parser_by_class(
-    cls: Type[gitlab.base.RESTObject],
-    sub_parser: _SubparserType,
+    cls: type[gitlab.base.RESTObject], sub_parser: _SubparserType
 ) -> None:
     mgr_cls_name = f"{cls.__name__}Manager"
     mgr_cls = getattr(gitlab.v4.objects, mgr_cls_name)
 
-    action_parsers: Dict[str, argparse.ArgumentParser] = {}
-    for action_name in ["list", "get", "create", "update", "delete"]:
+    action_parsers: dict[str, argparse.ArgumentParser] = {}
+    for action_name, help_text in [
+        ("list", "List the GitLab resources"),
+        ("get", "Get a GitLab resource"),
+        ("create", "Create a GitLab resource"),
+        ("update", "Update a GitLab resource"),
+        ("delete", "Delete a GitLab resource"),
+    ]:
         if not hasattr(mgr_cls, action_name):
             continue
 
         sub_parser_action = sub_parser.add_parser(
-            action_name, conflict_handler="resolve"
+            action_name, conflict_handler="resolve", help=help_text
         )
         action_parsers[action_name] = sub_parser_action
         sub_parser_action.add_argument("--sudo", required=False)
@@ -230,12 +245,25 @@ def _populate_sub_parser_by_class(
 
             sub_parser_action.add_argument("--page", required=False, type=int)
             sub_parser_action.add_argument("--per-page", required=False, type=int)
-            sub_parser_action.add_argument(
+            get_all_group = sub_parser_action.add_mutually_exclusive_group()
+            get_all_group.add_argument(
                 "--get-all",
                 required=False,
-                action="store_true",
+                action="store_const",
+                const=True,
+                default=None,
+                dest="get_all",
                 help="Return all items from the server, without pagination.",
             )
+            get_all_group.add_argument(
+                "--no-get-all",
+                required=False,
+                action="store_const",
+                const=False,
+                default=None,
+                dest="get_all",
+                help="Don't return all items from the server.",
+            )
 
         if action_name == "delete":
             if cls._id_attr is not None:
@@ -262,6 +290,10 @@ def _populate_sub_parser_by_class(
                 sub_parser_action.add_argument(
                     f"--{x.replace('_', '-')}", required=False
                 )
+            if mgr_cls._create_attrs.exclusive:
+                group = sub_parser_action.add_mutually_exclusive_group()
+                for x in mgr_cls._create_attrs.exclusive:
+                    group.add_argument(f"--{x.replace('_', '-')}")
 
         if action_name == "update":
             if cls._id_attr is not None:
@@ -280,14 +312,22 @@ def _populate_sub_parser_by_class(
                         f"--{x.replace('_', '-')}", required=False
                     )
 
+            if mgr_cls._update_attrs.exclusive:
+                group = sub_parser_action.add_mutually_exclusive_group()
+                for x in mgr_cls._update_attrs.exclusive:
+                    group.add_argument(f"--{x.replace('_', '-')}")
+
     if cls.__name__ in cli.custom_actions:
         name = cls.__name__
         for action_name in cli.custom_actions[name]:
+            custom_action = cli.custom_actions[name][action_name]
             # NOTE(jlvillal): If we put a function for the `default` value of
             # the `get` it will always get called, which will break things.
             action_parser = action_parsers.get(action_name)
             if action_parser is None:
-                sub_parser_action = sub_parser.add_parser(action_name)
+                sub_parser_action = sub_parser.add_parser(
+                    action_name, help=custom_action.help
+                )
             else:
                 sub_parser_action = action_parser
             # Get the attributes for URL/path construction
@@ -300,17 +340,16 @@ def _populate_sub_parser_by_class(
 
             # We need to get the object somehow
             if not issubclass(cls, gitlab.mixins.GetWithoutIdMixin):
-                if cls._id_attr is not None:
+                if cls._id_attr is not None and custom_action.requires_id:
                     id_attr = cls._id_attr.replace("_", "-")
                     sub_parser_action.add_argument(f"--{id_attr}", required=True)
 
-            required, optional, dummy = cli.custom_actions[name][action_name]
-            for x in required:
+            for x in custom_action.required:
                 if x != cls._id_attr:
                     sub_parser_action.add_argument(
                         f"--{x.replace('_', '-')}", required=True
                     )
-            for x in optional:
+            for x in custom_action.optional:
                 if x != cls._id_attr:
                     sub_parser_action.add_argument(
                         f"--{x.replace('_', '-')}", required=False
@@ -333,13 +372,13 @@ def _populate_sub_parser_by_class(
                     )
                 sub_parser_action.add_argument("--sudo", required=False)
 
-            required, optional, dummy = cli.custom_actions[name][action_name]
-            for x in required:
+            custom_action = cli.custom_actions[name][action_name]
+            for x in custom_action.required:
                 if x != cls._id_attr:
                     sub_parser_action.add_argument(
                         f"--{x.replace('_', '-')}", required=True
                     )
-            for x in optional:
+            for x in custom_action.optional:
                 if x != cls._id_attr:
                     sub_parser_action.add_argument(
                         f"--{x.replace('_', '-')}", required=False
@@ -355,18 +394,25 @@ def extend_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
     subparsers.required = True
 
     # populate argparse for all Gitlab Object
-    classes = set()
+    classes: set[type[gitlab.base.RESTObject]] = set()
     for cls in gitlab.v4.objects.__dict__.values():
         if not isinstance(cls, type):
             continue
         if issubclass(cls, gitlab.base.RESTManager):
-            if cls._obj_cls is not None:
-                classes.add(cls._obj_cls)
+            classes.add(cls._obj_cls)
 
     for cls in sorted(classes, key=operator.attrgetter("__name__")):
+        if cls is gitlab.base.RESTObject:
+            # Skip managers where _obj_cls is a plain RESTObject class
+            # Those managers do not actually manage any objects and
+            # can only be used to calls specific API paths.
+            continue
+
         arg_name = cli.cls_to_gitlab_resource(cls)
+        mgr_cls_name = f"{cls.__name__}Manager"
+        mgr_cls = getattr(gitlab.v4.objects, mgr_cls_name)
         object_group = subparsers.add_parser(
-            arg_name, formatter_class=cli.VerticalHelpFormatter
+            arg_name, help=f"API endpoint: {mgr_cls._path}"
         )
 
         object_subparsers = object_group.add_subparsers(
@@ -381,8 +427,8 @@ def extend_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
 
 
 def get_dict(
-    obj: Union[str, Dict[str, Any], gitlab.base.RESTObject], fields: List[str]
-) -> Union[str, Dict[str, Any]]:
+    obj: str | dict[str, Any] | gitlab.base.RESTObject, fields: list[str]
+) -> str | dict[str, Any]:
     if not isinstance(obj, gitlab.base.RESTObject):
         return obj
 
@@ -393,25 +439,21 @@ def get_dict(
 
 class JSONPrinter:
     @staticmethod
-    def display(d: Union[str, Dict[str, Any]], **_kwargs: Any) -> None:
-        import json  # noqa
-
+    def display(d: str | dict[str, Any], **_kwargs: Any) -> None:
         print(json.dumps(d))
 
     @staticmethod
     def display_list(
-        data: List[Union[str, Dict[str, Any], gitlab.base.RESTObject]],
-        fields: List[str],
+        data: list[str | dict[str, Any] | gitlab.base.RESTObject],
+        fields: list[str],
         **_kwargs: Any,
     ) -> None:
-        import json  # noqa
-
         print(json.dumps([get_dict(obj, fields) for obj in data]))
 
 
 class YAMLPrinter:
     @staticmethod
-    def display(d: Union[str, Dict[str, Any]], **_kwargs: Any) -> None:
+    def display(d: str | dict[str, Any], **_kwargs: Any) -> None:
         try:
             import yaml  # noqa
 
@@ -425,8 +467,8 @@ class YAMLPrinter:
 
     @staticmethod
     def display_list(
-        data: List[Union[str, Dict[str, Any], gitlab.base.RESTObject]],
-        fields: List[str],
+        data: list[str | dict[str, Any] | gitlab.base.RESTObject],
+        fields: list[str],
         **_kwargs: Any,
     ) -> None:
         try:
@@ -446,14 +488,14 @@ class YAMLPrinter:
 
 
 class LegacyPrinter:
-    def display(self, _d: Union[str, Dict[str, Any]], **kwargs: Any) -> None:
+    def display(self, _d: str | dict[str, Any], **kwargs: Any) -> None:
         verbose = kwargs.get("verbose", False)
         padding = kwargs.get("padding", 0)
-        obj: Optional[Union[Dict[str, Any], gitlab.base.RESTObject]] = kwargs.get("obj")
+        obj: dict[str, Any] | gitlab.base.RESTObject | None = kwargs.get("obj")
         if TYPE_CHECKING:
             assert obj is not None
 
-        def display_dict(d: Dict[str, Any], padding: int) -> None:
+        def display_dict(d: dict[str, Any], padding: int) -> None:
             for k in sorted(d.keys()):
                 v = d[k]
                 if isinstance(v, dict):
@@ -507,10 +549,7 @@ class LegacyPrinter:
         )
 
     def display_list(
-        self,
-        data: List[Union[str, gitlab.base.RESTObject]],
-        fields: List[str],
-        **kwargs: Any,
+        self, data: list[str | gitlab.base.RESTObject], fields: list[str], **kwargs: Any
     ) -> None:
         verbose = kwargs.get("verbose", False)
         for obj in data:
@@ -521,9 +560,7 @@ class LegacyPrinter:
             print("")
 
 
-PRINTERS: Dict[
-    str, Union[Type[JSONPrinter], Type[LegacyPrinter], Type[YAMLPrinter]]
-] = {
+PRINTERS: dict[str, type[JSONPrinter] | type[LegacyPrinter] | type[YAMLPrinter]] = {
     "json": JSONPrinter,
     "legacy": LegacyPrinter,
     "yaml": YAMLPrinter,
@@ -534,10 +571,10 @@ def run(
     gl: gitlab.Gitlab,
     gitlab_resource: str,
     resource_action: str,
-    args: Dict[str, Any],
+    args: dict[str, Any],
     verbose: bool,
     output: str,
-    fields: List[str],
+    fields: list[str],
 ) -> None:
     g_cli = GitlabCLI(
         gl=gl,
@@ -547,7 +584,7 @@ def run(
     )
     data = g_cli.run()
 
-    printer: Union[JSONPrinter, LegacyPrinter, YAMLPrinter] = PRINTERS[output]()
+    printer: JSONPrinter | LegacyPrinter | YAMLPrinter = PRINTERS[output]()
 
     if isinstance(data, dict):
         printer.display(data, verbose=True, obj=data)
