@@ -51,7 +51,14 @@ class ProjectCiLintManager(
     _path = "/projects/{project_id}/ci/lint"
     _obj_cls = ProjectCiLint
     _from_parent_attrs = {"project_id": "id"}
-    _optional_get_attrs = ("dry_run", "include_jobs", "ref")
+    _optional_get_attrs = (
+        "content_ref",
+        "dry_run",
+        "dry_run_ref",
+        "include_jobs",
+        "ref",
+    )
+
     _create_attrs = RequiredOptional(
         required=("content",), optional=("dry_run", "include_jobs", "ref")
     )
